-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

package.path = "lua/vehicle/?.lua;lua/vehicle/jbeam/?.lua;lua/common/?.lua;lua/?.lua;?.lua"
require('compatibility')

log = function(...) Lua:log(...) end
print = function(...) Lua:log('A', "print", tostring(...)) end

require('utils')
require('mathlib')

math.randomseed(os.time())

-- improve stacktraces
local STP = require "StackTracePlus"
debug.traceback = STP.stacktrace

perf = require("perf")
settings = require('simplesettings')

vehicleMode = 0
lastDt = 1/20

loadingTimes = {} -- global intentionally, used all over the place

local engineShiftUpRPM = 0

local function nop() end
log_jbeam = nop -- intentionally global

local loaded = false

-- step functions
function physicsStep(dts)
    drivetrain.updateWheels(dts)
    sensors.update(dts)
    thrusters.update()
    hydros.update(dts)
end

-- This is called in the global scope, so it is safe to do things that contact things outside the vehicle
function graphicsStep(dt)
    beamstate.updateGFX()
    bullettime.update(dt)
end

-- This is called in the local scope, so it is NOT safe to do things that contact things outside the vehicle
-- It is called right after graphicsStep
function graphicsStepLocal(dt)
    if not loaded then return end
    lastDt = dt
    sensors.updateGFX(dt) -- must be before input and ai
    mapmgr.sendTracking() -- must be before ai.updateGFX
    ai.updateGFX(dt) -- must be before input
    input.update(dt) -- must be as early as possible
    electrics.update(dt)
    extensions.hook('updateGFX', dt)
    drivetrain.updateEngine(dt)
    sounds.update(dt)
    hydros.updateGFX(dt)
    thrusters.updateGFX()
    if vehicleMode == 1 then
        recovery.updateGFX(dt)
        perf.update()
        streams.updateGFX()
        gui.updateGFX(dt)
        guihooks.updateGFX()
        if settings and settings.getValue('rpmLedsEnabled') == true then
            -- void playRPMLeds(const float currentRPM, const float rpmFirstLedTurnsOn, const float rpmRedLine)
            --print('playRPMLeds| currentRPM=' .. tostring(drivetrain.rpm) .. ' , rpmFirstLedTurnsOn=' .. tostring(v.data.engine.shiftUpRPM * 0.8) .. ' , rpmRedLine =' .. tostring(v.data.engine.shiftUpRPM))
            obj:playRPMLeds(drivetrain.rpm, engineShiftUpRPM * 0.8, engineShiftUpRPM)
        end
    end
    drivetrain.updateWheelsGFX(dt)
    mapmgr.updateGFX()
    scenario.updateGFX()
    props.update()
    fire.updateGFX(dt)
end

-- debug rendering
function debugDraw(x, y, z)
    local focusPos = float3(0,0,0)
    if x and y and z then focusPos = float3(x,y,z) end
    bdebug.debugDraw(focusPos)
    ai.debugDraw(focusPos)
end

function enableCallbacks()
    -- this enables the most important frame callbacks
    -- in case of an error, they still point to nop and thus do not spam the console
    -- TODO: FIXME
end

function initSystems()
    material.init()
    drivetrain.init()
    sensors.reset()
    beamstate.init()
    thrusters.init()
    hydros.init()
    input.init()
    sounds.init()
    props.init()
    electrics.init()
    recovery.init()
    bdebug.init()
    scenario.init()
    sensors.init()
    fire.init()

    gui.message("", 0, "^damage\\.") -- clear damage messages on vehicle restart
    extensions.hook('onInit')
end

function init(path, partConfigData)
    local hp1 = HighPerfTimer()
    local hp2 = HighPerfTimer()

    --perf.enable(1)

    if not obj then log('W', "default.init", "Error getting main object: unable to spawn") ; return end
    log('D', "default.init", "spawning vehicle "..tostring(path))

    -- we change the lookup path here, so it prefers the vehicle lua
    package.path = path .. "/lua/?.lua;".. package.path
    extensions = require("extensions")
    extensions.loadModules(path.."/lua")
    extensions.addModulePath("lua/vehicle/extensions/")
    extensions.loadModules("lua/vehicle/extensions/auto")

    drivetrain = require("drivetrain")
    sounds     = require("sounds")
    bdebug     = require("bdebug")
    input      = require("input")
    props      = require("props")

    particlefilter = require("particlefilter")
    particles  = require("particles")
    material   = require("material")
    v          = require("jbeam_main")
    electrics  = require("electrics")
    json       = require("json")
    beamstate  = require("beamstate")
    sensors    = require("sensors")
    bullettime = require("bullettime")
    thrusters  = require("thrusters")
    hydros     = require("hydros")
    gui        = require("gui") -- do not change its name, the GUI callback will break otherwise
    partmgmt   = require("partmgmt") -- do not change its name, the GUI callback will break otherwise
    streams    = require("guistreams")
    guihooks   = require("guihooks")
    ai         = require("ai")
    recovery   = require("recovery")
    mapmgr     = require("mapmgr")
    scenario   = require("scenario")
    fire       = require("fire")

    table.insert(loadingTimes, {'0 startup', hp1:stopAndReset()})

    -- care about the config before pushing to the physics
    if type(partConfigData) == 'string' and string.len(partConfigData) > 0 then
        local firstChar = string.sub(partConfigData,1,1)
        if firstChar == '{' or firstChar == '[' then
            log('D', "default.init", "  using partconfig data "..tostring(partConfigData)..' - path:'..tostring(path) )
            partmgmt.setConfig(unserialize(partConfigData), false)
        else
            log('D', "default.init", "  using partconfig filename "..tostring(partConfigData)..' - path:'..tostring(path) )
            partmgmt.load(partConfigData, false)
        end
    end

    -- enable/disable Jbeam logging depending on the devmode
    if settings and settings.getValue('devMode') == true then
      log_jbeam = log
    end
    
    -- experimental: vehicle file caching
    --[[
    local mp = require 'MessagePack'
    local vCacheFilename = 'vehicleCache.bin'
    if FS:fileExists(vCacheFilename) then
        local f = io.open(vCacheFilename, 'r')
        if f then
            v.vehicles = mp.unpack(f:read('*a'))
            f:close()
        end
    end
    --]]
    
    --if v.vehicles == nil then
        -- load jbeam files
        if not v.loadDirectories({'vehicles/common/', path}) then -- important: the real vehicle dir is always last, all additional paths come first
            return
        end
        table.insert(loadingTimes, {'1.X.X.X loadDirectories (sum)', hp1:stopAndReset()})

        -- experimental: vehicle file caching
        --[[
        local mp = require 'MessagePack'
        local f = io.open(vCacheFilename, 'w')
        if f then
            f:write(mp.pack(v.vehicles))
            f:close()
        end
        table.insert(loadingTimes, {'2. messagepack serialize', hp1:stopAndReset()})
        --]]
    --end

    -- you can change the data in here before it gets submitted to the physics

    -- submit to physics
    hp1:reset()
    v.pushToPhysics(obj)
    table.insert(loadingTimes, {'2.XX pushToPhysics (sum)', hp1:stopAndReset()})

    if v.data == nil then
        v.data = {}
    end

    initSystems()
    table.insert(loadingTimes, {'3 initSystems', hp1:stopAndReset()})

    -- temporary tire mark setting
    obj.slipTireMarkThreshold = 15

    -- if we got to here, lets enable the callbacks and hope for the best
    enableCallbacks()

    if v.data.engine ~= nil then
        engineShiftUpRPM = v.data.engine.shiftUpRPM or 0
    end


    -- update gravity and other things
    obj:queueGameEngineLua( 'vehicleSpawned('..tostring(obj:getID())..')' )

    if settings and settings.getValue('outgaugeEnabled') == true then
        extensions.loadModule("outgauge")
    end

    local totalTime = hp2:stopAndReset()
    table.insert(loadingTimes, {'4.X.X.X total (sum)', totalTime})

    log('D', "default.init", "init done - vehicle loading took "..totalTime..' ms')

    --perf.disable()
    --perf.saveDataToCSV('vehicle_boottime_perf.csv')

    -- dump boot time stats
    log('D', 'default.init', 'Vehicle boot time stats:')
    for _, t in pairs(loadingTimes) do
        log('D', 'default.init', '  * ' .. rpad(t[1], 34, ' ') .. ' = ' .. lpad(string.format('%5.3f', t[2]), 8, ' ') .. ' ms')
    end
    -- dump them into a file as well
    --[[
    local f = io.open('vehicle_boottime.csv', "w")
    if f then
        table.sort(loadingTimes, function(a, b) return a[2] > b[2] end)
        f:write(";name, time in ms\n")
        for _,vv in ipairs(loadingTimes) do
            f:write(vv[1] .. ', ' .. tostring(vv[2]) .. "\n")
        end
        f:close()
    end
    --]]


    loaded = true
end

-- various callbacks
function beamBroken(id, energy)
    beamstate.beamBroken(id, energy)
    --bullettime.slowMotion(0.05, 20) -- 20% of realtime
end

function beamDeformed(id, ratio)
    beamstate.beamDeformed(id, ratio)
end

-- called when everything was cleared
function vehicleDestroy()
    --log('D', "default.vehicleDestroy", "vehicleDestroy()")
    -- when the vehicle gets unloaded, remove all sounds
    sounds.destroy()
    material.destroy()
end

-- called when the user pressed I
function vehicleResetted(retainDebug)
    local hp1 = HighPerfTimer()
    --log('D', "default.vehicleResetted", "vehicleResetted()")
    electrics.reset()
    drivetrain.reset()
    props.reset()
    sensors.reset()
    if not retainDebug then
        bdebug.reset()
    end
    bullettime.reset()
    beamstate.reset()
    thrusters.reset()
    input.reset()
    hydros.reset()
    material.reset()
    guihooks.reset()
    extensions.hook('onReset', retainDebug)
    ai.reset()
    partmgmt.vehicleResetted()
    mapmgr.reset()
    scenario.reset()
    fire.reset()

    gui.message("", 0, "^damage\\.") -- clear damage messages on vehicle restart
    obj:queueGameEngineLua("extensions.hook('onVehicleResetted'," .. obj:getID() .. ")")
    loadingTimes['4_vehicleResetted'] = hp1:stopAndReset()
end

function nodeCollision(p)
    drivetrain.updateWheelSlip(p)
    fire.nodeCollision(p)
    particlefilter.nodeCollision(p)
    --sounds.nodeCollision(p)
end

function activated(mode)
    vehicleMode = mode
    if vehicleMode == 1 then mapmgr.enableTracking() end
    if vehicleMode == 0 then mapmgr.disableTracking() end
    --[[
    if mode == 1 then
        log('D', "default.activated", "yay, we got active")
    else
        log('D', "default.activated", "deactivated :(")
    end
    ]]--

    bdebug.activated(mode)
    ai.stateChanged()
    extensions.hook('activated', mode)
    guihooks.trigger("VehicleFocusChanged", {id = obj:getID(), mode = mode})
end

function instabilityDetected()
    -- enable break debug, so we see what beams broke
    local msg = "instability detected for vehicle " .. tostring(v.vehicleDirectory) .. ": resetting the vehicle and disabling the physics. "
    log('E', "default.instabilityDetected", msg)

    ui_message(msg, 10, 'instability')

    BeamEngine:setEnabled(false)
    obj:requestReset(RESET_PHYSICS)

   --if settings and settings.getValue('disableDynamicCollision') == true then
   --  Settings.dynamicCollision = false
   --end
end

function exportPersistentData()
    local s = serializeModules()
    --log('D', "default.exportPersistentData", s)
    obj:setPersistentData(s)
end

function importPersistentData(s)
    --log('D', "default.importPersistentData", s)
    -- deserialize extensions first, so the extensions are loaded before they are trying to get deserialized
    deserializeModules(unserialize(s), 'extensions')
    deserializeModules(unserialize(s))
end

function guiUpdate(d)
    --log('D', "default.guiUpdate", "guiUpdate()")
    gui.gotUpdate(d)
end

-- used by the GE camera code
function requestCameraConfig()
  if not cameraData then cameraData = {} end
  obj:queueGameEngineLua("extensions.hook('onVehicleCameraConfigChanged'," .. obj:getID() .. "," .. serialize(cameraData) .. ")")
end
